// +build windows

package main

import (

	"os"
	"fmt"
	"log"
	"time"
	"flag"
	"path/filepath"

	"github.com/revel/revel"

	"golang.org/x/sys/windows/svc"
	"golang.org/x/sys/windows/svc/mgr"
	"golang.org/x/sys/windows/svc/debug"
	"golang.org/x/sys/windows/svc/eventlog"
)

const (

	svcNameDefault = `sql4dialer.web`
)

var (

	serviceName  *string = flag.String("service", svcNameDefault, "Optional service name")
	uninstallSvc *bool = flag.Bool("uninstall", false, "Remove as a service")
	installSvc   *bool = flag.Bool("install", false, "Install as a service, with optional service name")
)



func main() {

	flag.Parse()

	isIntSess, err := svc.IsAnInteractiveSession()
	if err != nil {
		log.Fatalf("failed to determine if we are running in an interactive session: %v", err)
	}
	if !isIntSess {
		serviceRun(*serviceName, false)
		return
	}

	if *installSvc {
		err := serviceCreate(*serviceName, "SQL4Dialer Websrv")
		if err != nil {
			log.Fatalf("failed to create service(%s): %v", *serviceName, err)
		}
		return
	}

	if *uninstallSvc {
		err := serviceDelete(*serviceName)
		if err != nil {
			log.Fatalf("failed to delete service(%s): %v", *serviceName, err)
		}
		return
	}

	Run()
}



var elog debug.Log

type winsvc struct{}

func (m *winsvc) Execute(args []string, r <-chan svc.ChangeRequest, changes chan<- svc.Status) (ssec bool, errno uint32) {
	// fmt.Printf("svc.Execute(%#v, %#v, %#v)\n", args, r, changes)

	const cmdsAccepted = svc.AcceptStop | svc.AcceptShutdown // | svc.AcceptPauseAndContinue
	changes <- svc.Status{State: svc.StartPending}
	
	if !flag.Parsed() {
		flag.Parse()
	}

	revel.OnAppStart(func() {
		changes <- svc.Status{State: svc.Running, Accepts: cmdsAccepted}
	})

	go Run()
	time.Sleep(time.Millisecond *100)

 loop:
	for {
		select {

		case c := <-r:
			switch c.Cmd {
			case svc.Interrogate:
			{
				changes <- c.CurrentStatus
				// Testing deadlock from https://code.google.com/p/winsvc/issues/detail?id=4
				time.Sleep(time.Millisecond *100)
				changes <- c.CurrentStatus
			}
			case svc.Stop, svc.Shutdown:
			{
				break loop
			}
			default:
			{
				elog.Error(1, fmt.Sprintf("unexpected control request #%d", c))
			}}
		}
	}
	changes <- svc.Status{State: svc.StopPending}
	return
}

func serviceRun(serviceName string, isDebug bool) {
	var err error
	if isDebug {
		elog = debug.New(serviceName)
	} else {
		elog, err = eventlog.Open(serviceName)
		if err != nil {
			return
		}
	}
	defer elog.Close()

	elog.Info(1, fmt.Sprintf("starting %s service", serviceName))
	run := svc.Run
	if isDebug {
		run = debug.Run
	}
	err = run(serviceName, &winsvc{})
	if err != nil {
		elog.Error(1, fmt.Sprintf("%s service failed: %v", serviceName, err))
		return
	}
	elog.Info(1, fmt.Sprintf("%s service stopped", serviceName))
}

func exePath() (string, error) {
	prog := os.Args[0]
	p, err := filepath.Abs(prog)
	if err != nil {
		return "", err
	}
	fi, err := os.Stat(p)
	if err == nil {
		if !fi.Mode().IsDir() {
			return p, nil
		}
		err = fmt.Errorf("%s is directory", p)
	}
	if filepath.Ext(p) == "" {
		p += ".exe"
		fi, err := os.Stat(p)
		if err == nil {
			if !fi.Mode().IsDir() {
				return p, nil
			}
			err = fmt.Errorf("%s is directory", p)
		}
	}
	return "", err
}

func serviceCreate(serviceName, desc string) error {
	exepath, err := exePath()
	if err != nil {
		return err
	}
	m, err := mgr.Connect()
	if err != nil {
		return err
	}
	defer m.Disconnect()
	s, err := m.OpenService(serviceName)
	if err == nil {
		s.Close()
		return fmt.Errorf("service %s already exists", serviceName)
	}
	args := []string {
		fmt.Sprintf("-srcPath=%s", filepath.Join(filepath.Dir(exepath), "src")),
		fmt.Sprintf("-importPath=%s", "webitel.com/sql4dialer"),
		fmt.Sprintf("-runMode=%s", "prod"),
	}
	if *port != 0 {
		args = append(args, fmt.Sprintf("-port=%d", *port))
	}
	if serviceName != svcNameDefault {
		args = append(args, fmt.Sprintf("-service=%s", serviceName))
	}
	s, err = m.CreateService(serviceName, exepath, mgr.Config{DisplayName: desc}, args...)
	if err != nil {
		return err
	}
	defer s.Close()
	// err = eventlog.InstallAsEventCreate(serviceName, eventlog.Error|eventlog.Warning|eventlog.Info)
	// if err != nil {
	// 	s.Delete()
	// 	return fmt.Errorf("SetupEventLogSource() failed: %s", err)
	// }
	return nil
}

func serviceDelete(serviceName string) error {
	m, err := mgr.Connect()
	if err != nil {
		return err
	}
	defer m.Disconnect()
	s, err := m.OpenService(serviceName)
	if err != nil {
		return fmt.Errorf("service %s is not installed", serviceName)
	}
	defer s.Close()
	err = s.Delete()
	if err != nil {
		return err
	}
	// err = eventlog.Remove(serviceName)
	// if err != nil {
	// 	return fmt.Errorf("RemoveEventLogSource() failed: %s", err)
	// }
	return nil
}